//-----------------------------------------------------------------------------
// File: Pick.cpp
//
// Desc: Example code showing how to do picking in D3D.
//
// Copyright (c) 1997-2000 Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 8.0 SDK Delphi adaptation by Alexey Barkovoy *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  Modified: 18-Feb-2001                                                     *
 *                                                                            *
 *  Lastest version can be downloaded from:                                   *
 *     http://groups.yahoo.com/group/JEDI-DirectXExamples                     *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)

unit PickUnit;

{$A+}  // Word alignment data
{$OA+} // Objects and structures align
{$Z4}  // Set minimum size of enumerated type to 4

interface

uses
  Windows, Messages, D3D, D3D8, D3DX8, D3DApp, D3DFont, D3DFile, D3DUtil, DXUtil;

type
  D3DVERTEX = packed record
    p: TD3DXVector3;
    n: TD3DXVector3;
    tu, tv: Single;
  end;

const
  D3DFVF_VERTEX = D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_TEX1;

type
//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the
//       generic functionality needed in all Direct3D samples. CMyD3DApplication
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
  CMyD3DApplication = ^TMyD3DApplication;
  TMyD3DApplication = object(TD3DApplication)
  private
    m_pFont:    CD3DFont;                      // Font for drawing text
    m_pObject:  CD3DMesh;                      // Object to render
    m_dwNumPickedTriangles: DWord;             // Picking information
    m_fPickT, m_fPickU, m_fPickV: Single;      // Picking results
    m_pVB: IDirect3DVertexBuffer8;             // VB for picked triangles
    m_pCursorBitmap: IDirect3DSurface8;        // Cursor image

    // Internal member functions
    function Pick: HResult;
    function IntersectTriangle(const orig, dir: TD3DXVector3;
                               const v0, v1, v2: TD3DXVector3;
                               out t, u, v: Single): BOOL;

    function OneTimeSceneInit: HResult; virtual;
    function InitDeviceObjects: HResult; virtual;
    function RestoreDeviceObjects: HResult; virtual;
    function InvalidateDeviceObjects: HResult; virtual;
    function DeleteDeviceObjects: HResult; virtual;
    function Render: HResult; virtual;
    function FrameMove: HResult; virtual;
    function FinalCleanup: HResult; virtual;
    function ConfirmDevice(var pCaps: TD3DCaps8; dwBehavior: DWord; Format: TD3DFormat): HResult; virtual;

  public
    constructor Create;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM ): LRESULT; virtual;
  end;

implementation

uses
  Strings;


//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor TMyD3DApplication.Create;
begin
  inherited Create(0);
  m_strWindowTitle    := 'Pick: D3D Picking Sample';
  m_bUseDepthBuffer   := TRUE;
  m_bShowCursorWhenFullscreen := TRUE;

  New(m_pFont,    Create('Arial', 12, D3DFONT_BOLD));
  New(m_pObject,  CreateNoName);
  m_pCursorBitmap   := nil;
  m_pVB             := nil;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function TMyD3DApplication.OneTimeSceneInit;
begin
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Pick()
// Desc: Checks if mouse point hits geometry
//       the scene.
//-----------------------------------------------------------------------------
function TMyD3DApplication.Pick;
type
  TD3DVertexArray = array [0..16383] of D3DVERTEX;
  PWordArray = ^TWordArray;
  TWordArray = array[0..16383] of Word;
var
  vPickRayDir: TD3DXVector3;
  vPickRayOrig: TD3DXVector3;

  matProj: TD3DXMatrix;
  ptCursor: TPoint;
  v: TD3DXVector3;

  pVB: IDirect3DVertexBuffer8;
  pIB: IDirect3DIndexBuffer8;

  // dwFVF: DWord;
  matView, m: TD3DXMatrix;

  pVertices: ^TD3DVertexArray;
  pIndices: ^TWordArray;
  dwNumFaces: DWord;

  i: DWord;
  v0, v1, v2: TD3DXVector3;
  pV: ^TD3DVertexArray;
begin
  m_dwNumPickedTriangles:= 0;

  // Get the pick ray from the mouse position
  if GetCapture <> 0 then
  begin
    m_pd3dDevice.GetTransform(D3DTS_PROJECTION, matProj);

    GetCursorPos(ptCursor);
    ScreenToClient(m_hWnd, ptCursor);

    // Compute the vector of the pick ray in screen space
    v.x :=  ( ( ( 2.0 * ptCursor.x ) / m_d3dsdBackBuffer.Width  ) - 1 ) / matProj._11;
    v.y := -( ( ( 2.0 * ptCursor.y ) / m_d3dsdBackBuffer.Height ) - 1 ) / matProj._22;
    v.z :=  1.0;

    // Get the inverse view matrix
    m_pd3dDevice.GetTransform(D3DTS_VIEW, matView );
    D3DXMatrixInverse(m, nil, matView);

    // Transform the screen space pick ray into 3D space
    vPickRayDir.x  := v.x*m._11 + v.y*m._21 + v.z*m._31;
    vPickRayDir.y  := v.x*m._12 + v.y*m._22 + v.z*m._32;
    vPickRayDir.z  := v.x*m._13 + v.y*m._23 + v.z*m._33;
    vPickRayOrig.x := m._41;
    vPickRayOrig.y := m._42;
    vPickRayOrig.z := m._43;
  end;

  // Get the picked triangle
  if GetCapture <> 0 then
  begin
    m_pObject^.GetLocalMesh.GetVertexBuffer(pVB);
    m_pObject^.GetLocalMesh.GetIndexBuffer(pIB);

    // dwFVF := m_pObject.GetLocalMesh.GetFVF; - it's not used anyway

    dwNumFaces:= m_pObject^.GetLocalMesh.GetNumFaces;
    pIB.Lock(0,0, PByte(pIndices), 0);
    pVB.Lock(0,0, PByte(pVertices), 0);

    for i:= 0 to dwNumFaces - 1 do
    begin
      v0 := pVertices^[pIndices^[3*i+0]].p;
      v1 := pVertices^[pIndices^[3*i+1]].p;
      v2 := pVertices^[pIndices^[3*i+2]].p;

      // Check if the pick ray passes through this point
      if IntersectTriangle(vPickRayOrig, vPickRayDir, v0, v1, v2,
                           m_fPickT, m_fPickU, m_fPickV) then
      begin
        m_pVB.Lock( 0, 0, PByte(pV), 0 );
        pV^[3*m_dwNumPickedTriangles+0] := pVertices^[pIndices^[3*i+0]];
        pV^[3*m_dwNumPickedTriangles+1] := pVertices^[pIndices^[3*i+1]];
        pV^[3*m_dwNumPickedTriangles+2] := pVertices^[pIndices^[3*i+2]];
        m_pVB.Unlock();

        Inc(m_dwNumPickedTriangles);
      end;
    end;

    pVB.Unlock;
    pIB.Unlock;

    pVB:= nil;
    pIB:= nil;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function TMyD3DApplication.FrameMove;
var
  vFromPt:      TD3DXVector3;
  vLookatPt:    TD3DXVector3;
  vUpVec:       TD3DXVector3;

  matView: TD3DXMatrix;
begin
  // Rotate the camera about the y-axis
  vFromPt   := D3DXVector3Zero;
  vLookatPt := D3DXVector3Zero;
  vUpVec    := D3DXVector3(0.0, 1.0, 0.0);
  vFromPt.x := -Cos(m_fTime/3.0) * 4.0;
  vFromPt.y := 1.0;
  vFromPt.z :=  Sin(m_fTime/3.0) * 4.0;

  D3DXMatrixLookAtLH(matView, vFromPt, vLookatPt, vUpVec);
  m_pd3dDevice.SetTransform(D3DTS_VIEW, matView);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function TMyD3DApplication.Render;
var
  ptCursor: TPoint;

  strBuffer: array [0..89] of Char;
  S: String;
begin
  // Set up the cursor
  GetCursorPos(ptCursor);
  ScreenToClient(m_hWnd, ptCursor);
  m_pd3dDevice.SetCursorPosition(ptCursor.x, ptCursor.y, 0);

  // Check for picked triangles
  Pick;

  // Clear the viewport
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
                     $000000ff, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    // Set render mode to lit, solid triangles
    m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
    m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, Integer(TRUE));

    // If a triangle is picked, draw it
    if m_dwNumPickedTriangles <> 0 then
    begin
      // Draw the picked triangle
      m_pd3dDevice.SetVertexShader( D3DFVF_VERTEX );
      m_pd3dDevice.SetStreamSource( 0, m_pVB, sizeof(D3DVERTEX));
      m_pd3dDevice.DrawPrimitive( D3DPT_TRIANGLELIST,
                                   0, m_dwNumPickedTriangles);

      // Set render mode to unlit, wireframe triangles
      m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);
      m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, Integer(FALSE));
    end;

    m_pObject^.Render(m_pd3dDevice, True, True);

    // Output statistics
    m_pFont^.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats,  0);
    m_pFont^.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats, 0);

    // Output text
    if (m_dwNumPickedTriangles <> 0) then
    begin
      S:= 't=' + Fix(m_fPickT, 0, 2) + ', u=' + Fix(m_fPickU, 0, 2) + ', v=' + Fix(m_fPickV, 0, 2);
      StrPCopy(strBuffer, S);
    end
    else
      strBuffer:= 'Use mouse to pick a polygon' + #0;

    m_pFont^.DrawText(2, 40, D3DCOLOR_ARGB(255,255,255,0), strBuffer, 0);

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function TMyD3DApplication.InitDeviceObjects;
const
  MeshName: PChar = 'Tiger.x';
var
  dwNumVertices: DWord;
begin
  // Restore the font
  m_pFont^.InitDeviceObjects(m_pd3dDevice);

  if FAILED(m_pObject^.Create_(m_pd3dDevice, MeshName)) then
  begin
    Result:= D3DAPPERR_MEDIANOTFOUND;
    Exit;
  end;

  m_pObject^.SetFVF(m_pd3dDevice, D3DFVF_VERTEX);

  // Create and fill the vertex buffer
  dwNumVertices:= m_pObject^.GetSysMemMesh.GetNumVertices;
  if FAILED(m_pd3dDevice.CreateVertexBuffer(dwNumVertices*sizeof(D3DVERTEX),
                                            D3DUSAGE_WRITEONLY, D3DFVF_VERTEX,
                                            D3DPOOL_MANAGED, m_pVB)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function TMyD3DApplication.RestoreDeviceObjects;
var
  matIdentity: TD3DXMatrix;
  matProj: TD3DXMatrix;
  fAspect: Single;
  mtrl: TD3DMaterial8;
  light: TD3DLight8;
begin
  m_pFont^.RestoreDeviceObjects;

  // Restore device-memory objects for the mesh
  m_pObject^.RestoreDeviceObjects(m_pd3dDevice);

  // Set up the textures
  m_pd3dDevice.SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
  m_pd3dDevice.SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
  m_pd3dDevice.SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_MODULATE );
  m_pd3dDevice.SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
  m_pd3dDevice.SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );

  // Set miscellaneous render states
  m_pd3dDevice.SetRenderState( D3DRS_DITHERENABLE,   Integer(FALSE) );
  m_pd3dDevice.SetRenderState( D3DRS_SPECULARENABLE, Integer(FALSE) );
  m_pd3dDevice.SetRenderState( D3DRS_ZENABLE,        Integer(TRUE) );
  m_pd3dDevice.SetRenderState( D3DRS_AMBIENT,        $00444444);

  // Set the world matrix
  D3DXMatrixIdentity(matIdentity);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matIdentity);

  // Set the projection matrix
  fAspect:= m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Setup a material
  D3DUtil_InitMaterial(mtrl, 1.0, 1.0, 1.0, 1.0);
  m_pd3dDevice.SetMaterial(mtrl);

  // Set up lighting states
  D3DUtil_InitLight(light, D3DLIGHT_DIRECTIONAL, 0.1, -1.0, 0.1);
  m_pd3dDevice.SetLight(0, light);
  m_pd3dDevice.LightEnable(0, TRUE);
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, Integer(TRUE));

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function TMyD3DApplication.InvalidateDeviceObjects;
begin
  m_pFont^.InvalidateDeviceObjects;
  m_pObject^.InvalidateDeviceObjects;

  SAFE_RELEASE(m_pCursorBitmap);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function TMyD3DApplication.DeleteDeviceObjects;
begin
  m_pFont^.DeleteDeviceObjects;
  m_pObject^.Destroy_;

  SAFE_RELEASE(m_pVB);
  SAFE_RELEASE(m_pCursorBitmap);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function TMyD3DApplication.FinalCleanup;
begin
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pObject);
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function TMyD3DApplication.
  ConfirmDevice(var pCaps: TD3DCaps8; dwBehavior: DWord; Format: TD3DFormat): HResult;
begin
  OutputDebugInfo('in Confirm Device');

  if dwBehavior and D3DCREATE_PUREDEVICE <> 0 then
  begin
    Result:= E_FAIL; // GetTransform doesn't work on PUREDEVICE
    Exit;
  end;

  // If this is a TnL device, make sure it supports directional lights
  if (dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING <> 0) or
     (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING <> 0) then
  begin
    if (pCaps.VertexProcessingCaps and D3DVTXPCAPS_DIRECTIONALLIGHTS = 0) then
    begin
      Result:= E_FAIL;
      Exit;
    end;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Overrrides the main WndProc, so the sample can do custom message
//       handling (e.g. processing mouse, keyboard, or menu commands).
//-----------------------------------------------------------------------------
function TMyD3DApplication.
  MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM ): LRESULT;
begin
  case uMsg of
    WM_LBUTTONDOWN: SetCapture( hWnd ); // User pressed left mouse button
    WM_LBUTTONUP:   ReleaseCapture;     // The user released the left mouse button
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

//-----------------------------------------------------------------------------
// Name: IntersectTriangle()
// Desc: Given a ray origin (orig) and direction (dir), and three vertices of
//       of a triangle, this function returns TRUE and the interpolated texture
//       coordinates if the ray intersects the triangle
//-----------------------------------------------------------------------------
function TMyD3DApplication.
  IntersectTriangle(const orig, dir: TD3DXVector3;
                    const v0, v1, v2: TD3DXVector3;
                    out t, u, v: Single): BOOL;
var
  edge1: TD3DXVector3;
  edge2: TD3DXVector3;

  pvec: TD3DXVector3;
  det: Single;
  tvec: TD3DXVector3;
  qvec: TD3DXVector3;
  fInvDet: Single;
begin
  // Find vectors for two edges sharing vert0
  D3DXVec3Subtract(edge1, v1, v0);
  D3DXVec3Subtract(edge2, v2, v0);

  // Begin calculating determinant - also used to calculate U parameter
  D3DXVec3Cross(pvec, dir, edge2);

  // If determinant is near zero, ray lies in plane of triangle
  det:= D3DXVec3Dot(edge1, pvec);
  if (det < 0.0001) then
  begin
    Result:= FALSE;
    Exit;
  end;

  // Calculate distance from vert0 to ray origin
  D3DXVec3Subtract(tvec, orig, v0);

  // Calculate U parameter and test bounds
  u:= D3DXVec3Dot(tvec, pvec);
  if(u < 0.0) or (u > det) then
  begin
    Result:= FALSE;
    Exit;
  end;

  // Prepare to test V parameter
  D3DXVec3Cross(qvec, tvec, edge1);

  // Calculate V parameter and test bounds
  v:= D3DXVec3Dot(dir, qvec);
  if (v < 0.0) or (u + v > det) then
  begin
    Result:= FALSE;
    Exit;
  end;

  // Calculate t, scale parameters, ray intersects triangle
  t := D3DXVec3Dot(edge2, qvec);
  fInvDet:= 1.0 / det;
  t:= t * fInvDet;
  u:= u * fInvDet;
  v:= v * fInvDet;

  Result:= TRUE;
end;

end.